/*
 * SnapShotDialog.java
 * Created on 2011/10/13
 *
 * Copyright (C) 2011-2013 Nippon Telegraph and Telephone Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

package jp.co.ntt.oss.heapstats.gui;

import java.awt.Color;
import java.awt.Component;
import java.awt.Paint;
import java.io.IOException;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.RowSorter.SortKey;
import javax.swing.SortOrder;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableModel;
import javax.swing.table.TableRowSorter;

import jp.co.ntt.oss.heapstats.container.MapKey;
import jp.co.ntt.oss.heapstats.container.ObjectData;
import jp.co.ntt.oss.heapstats.container.SnapShot;
import jp.co.ntt.oss.heapstats.container.SnapShotHeader;
import jp.co.ntt.oss.heapstats.gui.graph.GraphColorHelper;
import jp.co.ntt.oss.heapstats.gui.snapshot.SnapShotHelper;
import jp.co.ntt.oss.heapstats.gui.snapshot.SnapShotReader;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.PiePlot;
import org.jfree.data.general.DefaultPieDataset;

/**
 * extends {@link JDialog}<br>
 * <br>
 * Draw the snapshot information.
 */
public class SnapShotDialog extends JDialog {
    /** serialVersionUID. */
    private static final long serialVersionUID = 7336022709943040450L;

    /** Main Component. */
    private Box comp = Box.createVerticalBox();

    /** Class of higher rank. */
    private Set<String> ranking;

    /** object lists. */
    private JTable entries;

    /**
     * Creates a modeless dialog with the specified title and with the specified
     * owner dialog.
     *
     * @param owner the owner Dialog from which the dialog is displayed or null
     *        if this dialog has no owner.
     */
    public SnapShotDialog(final JFrame owner) {
        super(owner);

        setSize(GuiCommon.DEFAULT_DIALOG_WIDTH, GuiCommon.DEFAULT_DIALOG_HEIGHT);
        setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
        setModal(false);
        setTitle(GuiCommon.getResources().getString(
                GuiCommon.TAB_TITLE_SNAPSHOT));
        comp.setBorder(BorderFactory.createEmptyBorder(GuiCommon.BORDER_SIZE,
                GuiCommon.BORDER_SIZE, GuiCommon.BORDER_SIZE,
                GuiCommon.BORDER_SIZE));
        add(comp);
        setLocationByPlatform(true);
    }

    /**
     * This method does the drawing SnapShot Data.
     *
     * @param data a snapshot Information
     * @param search Browse Class List
     * @param filter Exclusion list of classes
     * @throws IOException Any of the usual Input/Output related exceptions.
     * @throws ClassNotFoundException Class of a serialized object cannot be
     *         found.
     */
    public final void draw(final List<Object> data, final Object[] search,
            final List<Pattern> filter) throws IOException,
            ClassNotFoundException {
        comp.setVisible(false);
        for (Component obj : comp.getComponents()) {
            comp.remove(obj);
        }

        ranking = new HashSet<String>();

        long offset = (Long) data.get(0);
        SnapShotReader snapShotReader = (SnapShotReader) data.get(1);

        // SnapShot Info
        String[] tableHeader = new String[] {"", ""};
        DefaultTableModel snapShotInfoModel = new DefaultTableModel(
                tableHeader, 0);
        JTable snapShotInfo = new JTable(snapShotInfoModel);

        // objects List
        SnapShotEntriesModel entriesModel = new SnapShotEntriesModel();
        entries = new JTable(entriesModel) {
            /** serialVersionUID. */
            private static final long serialVersionUID = -925317579655004381L;

            @Override
            public Component prepareRenderer(final TableCellRenderer tcr,
                    final int row, final int column) {
                Component retValue = super.prepareRenderer(tcr, row, column);

                String name = (String) getValueAt(row, 1);
                if (ranking.size() < GuiCommon.getRankLevel()) {
                    ranking.add(name);
                }
                retValue.setForeground(getForeground());
                Paint bgPaint = GraphColorHelper.getColor((String) getValueAt(
                        row, 1));
                if (bgPaint != null && column == 0 && ranking.contains(name)) {
                    retValue.setBackground((Color) bgPaint);
                } else {
                    retValue.setBackground(Color.WHITE);
                }
                return retValue;
            }

            @Override
            public boolean isCellEditable(final int row, final int column) {
                return false;
            }
        };

        entries.getColumnModel().getColumn(0)
                .setPreferredWidth(GuiCommon.COL_SIZE_COLOR);
        entries.getColumnModel().getColumn(1)
                .setPreferredWidth(GuiCommon.COL_SIZE_CLASSNAME);

        TableRowSorter<TableModel> sorter = new TableRowSorter<TableModel>(
                entriesModel);
        sorter.setSortable(0, false);
        entries.setRowSorter(sorter);
        entries.getTableHeader().setReorderingAllowed(false);

        snapShotInfoModel.setRowCount(0);
        entriesModel.setRowCount(0);

        SnapShot snapShot = snapShotReader.getSnapShot(offset);

        ObjectData[] elements = SnapShotHelper.getSortedArray(
                snapShot.getElements(), SnapShotHelper.SortItem.SIZE,
                SnapShotHelper.SortOrder.DESC);

        // Filter Start
        List<ObjectData> result = getObjectsList(elements, search, filter);

        // Sets the objects data.
        for (ObjectData elmt : result) {
            Object[] ary = new Object[4];
            ary[1] = elmt.getName();
            ary[2] = elmt.getCount();
            ary[3] = elmt.getTotalSize();
            entriesModel.addRow(ary);
        }

        ExpansionContractionPanel snapInfo = new ExpansionContractionPanel(
                GuiCommon.getResources().getString(GuiCommon.TITLE_SNAPSHOT));

        ExpansionContractionPanel pieChart = new ExpansionContractionPanel(
                GuiCommon.getResources().getString(GuiCommon.TITLE_PIECHART));

        ExpansionContractionPanel objList = new ExpansionContractionPanel(
                GuiCommon.getResources().getString(GuiCommon.TITLE_OBJECTLIST));

        JScrollPane entryScroll = new JScrollPane();
        entryScroll.getViewport().setView(entries);
        objList.setComponent(entryScroll, true);

        ArrayList<SortKey> sortKeys = new ArrayList<SortKey>();
        sortKeys.add(new SortKey(3, SortOrder.DESCENDING));
        entries.getRowSorter().setSortKeys(sortKeys);

        addComp(snapInfo);
        addComp(pieChart);
        addComp(objList);

        setSnapShotInfo(snapShot, snapShotInfoModel);

        snapInfo.setComponent(snapShotInfo, false);

        // draw pie chart
        if ((search != null && search.length > 0)
                || (filter != null && filter.size() > 0)) {
            drawPieChart(snapShot, result.toArray(), pieChart, false);
        } else {
            drawPieChart(snapShot, result.toArray(), pieChart, true);
        }

        if (this.isVisible()) {
            comp.setVisible(true);
        }
    }

    /**
     * Clear the SnapShot Data.
     */
    public final void clear() {
        for (Component obj : comp.getComponents()) {
            comp.remove(obj);
        }
        comp.setVisible(false);
    }

    /**
     * Returns the object list of articles and then excluded.
     *
     * @param elements List of all objects
     * @param search Browse Class List
     * @param filter Exclusion list of classes
     * @return Returns the object list of articles and then excluded.
     */
    private List<ObjectData> getObjectsList(final ObjectData[] elements,
            final Object[] search, final List<Pattern> filter) {
        List<ObjectData> tmpList = new ArrayList<ObjectData>();
        List<ObjectData> result = new ArrayList<ObjectData>();

        // Start Filter
        for (ObjectData elmt : elements) {
            boolean match = false;
            if (filter != null) {
                for (Pattern ptn : filter) {
                    Matcher matcher = ptn.matcher(elmt.getName());
                    if (matcher.matches()) {
                        match = true;
                        break;
                    }
                }
            }

            if (!match) {
                tmpList.add(elmt);
            }
        }

        // Start Search
        for (ObjectData elmt : tmpList) {
            if (search != null && search.length > 0) {
                for (Object tmp : search) {
                    if (elmt.getName().equals(tmp)) {
                        result.add(elmt);
                    }
                }
            } else {
                result.add(elmt);
            }
        }
        return result;
    }

    /**
     * SnapShot Information on label data sets.
     *
     * @param snapShot SnapShot
     * @param snapShotInfoModel DefaultTableModel the Objects List
     */
    private void setSnapShotInfo(final SnapShot snapShot,
            final DefaultTableModel snapShotInfoModel) {
        SnapShotHeader hdr = snapShot.getHeader();

        String heapUsageStr = NumberFormat.getNumberInstance().format(
                snapShot.getHeapUsage() / GuiCommon.BINARY_PREFIX)
                + " KB";

        SimpleDateFormat sdf = new SimpleDateFormat(GuiCommon.DATE_FORMAT);
        String[][] infoData = {
                {
                        GuiCommon.getResources().getString(
                                GuiCommon.SNAPSHOTINFO_DATE),
                        sdf.format(hdr.getSnapShotDate())},
                {
                        GuiCommon.getResources().getString(
                                GuiCommon.SNAPSHOTINFO_ENTRYIES),
                        Long.toString(hdr.getNumEntries())},
                {
                        GuiCommon.getResources().getString(
                                GuiCommon.SNAPSHOTINFO_HEAP), heapUsageStr},
                {
                        GuiCommon.getResources().getString(
                                GuiCommon.SNAPSHOTINFO_CAUSE),
                        hdr.getCauseString()},
                {
                        GuiCommon.getResources().getString(
                                GuiCommon.SNAPSHOTINFO_GCCAUSE),
                        hdr.getGcCause()}};

        String[] infoHeader = {"", ""};
        snapShotInfoModel.setDataVector(infoData, infoHeader);
    }

    /**
     * Draw the large object heap usage pie chart.
     *
     * @param snapShot SnapShot
     * @param elements Objects Data
     * @param pieChart ExpansionContractionPanel is to set the chart.
     * @param others Other display flag. true is view.
     */
    private void drawPieChart(final SnapShot snapShot, final Object[] elements,
            final ExpansionContractionPanel pieChart, final boolean others) {

        DefaultPieDataset graphData = new DefaultPieDataset();
        long heapUsage = snapShot.getHeapUsage();
        for (int i = 0; (i < elements.length) && (i < GuiCommon.getRankLevel()); i++) {
            ObjectData odata = (ObjectData) elements[i];
            graphData.setValue(new MapKey(odata), odata.getTotalSize());
            heapUsage -= odata.getTotalSize();
        }

        if (others) {
            graphData.setValue(new MapKey(0, "Others"), heapUsage);
        }

        JFreeChart chart = ChartFactory.createPieChart("", graphData, false,
                true, true);
        chart.setBackgroundPaint(Color.WHITE);
        PiePlot plot = (PiePlot) chart.getPlot();
        plot.setBackgroundPaint(new Color(GuiCommon.getBgColor()));
        plot.setLabelGenerator(null);
        GraphColorHelper.adjustColor(graphData, plot);

        ChartPanel chartPanel = new ChartPanel(chart);
        pieChart.setComponent(chartPanel, true);
    }

    /**
     * Gets the main components that you want to SnapShotDialog.
     *
     * @return Main component
     */
    public final Box getComp() {
        return comp;
    }

    /**
     * Sets the main components that you want to SnapShotDialog.
     *
     * @param value Box
     */
    public final void setComp(final Box value) {
        comp = value;
        add(comp);
    }

    /**
     * The main component to be added to the specified component.
     *
     * @param value Component
     */
    private void addComp(final Component value) {
        comp.add(Box.createVerticalStrut(GuiCommon.BORDER_SIZE));
        comp.add(value);
        comp.revalidate();
    }
}
